<?php

namespace Drupal\product_gallery\Plugin\Field\FieldFormatter;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Core\Url;
use Drupal\image\Plugin\Field\FieldFormatter\ImageFormatterBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the 'product_gallery_field_formatter' formatter.
 *
 * @FieldFormatter(
 *   id = "product_gallery_field_formatter",
 *   label = @Translation("Product Gallery"),
 *   field_types = {
 *     "image"
 *   }
 * )
 */
class ProductGalleryFieldFormatter extends ImageFormatterBase implements ContainerFactoryPluginInterface {

  /**
   * The file URL generator.
   *
   * @var \Drupal\Core\File\FileUrlGeneratorInterface
   */
  protected $fileUrlGenerator;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Image entity storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $imageStyleStorage;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\Renderer
   */
  protected $renderer;

  /**
   * Constructs an ImageFormatter object.
   *
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the formatter is associated.
   * @param array $settings
   *   The formatter settings.
   * @param string $label
   *   The formatter label display setting.
   * @param string $view_mode
   *   The view mode.
   * @param array $third_party_settings
   *   Any third party settings.
   * @param object $current_user
   *   The current user.
   * @param \Drupal\Core\Entity\EntityStorageInterface $image_style_storage
   *   Image style storage.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer.
   * @param \Drupal\Core\File\FileUrlGeneratorInterface $fileUrlGenerator
   *   The file URL generator.
   */
  public function __construct(
    $plugin_id,
    $plugin_definition,
    FieldDefinitionInterface $field_definition,
    array $settings,
    $label,
    $view_mode,
    array $third_party_settings,
    object $current_user,
    EntityStorageInterface $image_style_storage,
    Renderer $renderer,
    FileUrlGeneratorInterface $fileUrlGenerator,
  ) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);
    $this->currentUser = $current_user;
    $this->imageStyleStorage = $image_style_storage;
    $this->renderer = $renderer;
    $this->fileUrlGenerator = $fileUrlGenerator;
  }

  /**
   * Creates an instance of the plugin.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   The container to pull out services used in the plugin.
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   *
   * @return static
   *   Returns an instance of this plugin.
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('current_user'),
      $container->get('entity_type.manager')->getStorage('image_style'),
      $container->get('renderer'),
      $container->get('file_url_generator')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'image_style_large' => '',
      'image_style_medium' => '',
      'image_style_medium' => '',
      'image_style_small' => '',
      'thumbnail_image_style' => '',
      'thumbnail_border_style' => 'round',
      'allow_thumbnail_overlap' => TRUE,
      'thumbnail_overlap_threshold' => '5',
      'enable_mouse_zoom' => TRUE,
      'default_zoom_scale' => '2',
      'show_magnifier' => TRUE,
      'image_class' => '',
      'image_wrapper_class' => '',
      'container_width' => '600',
      'container_width_unit' => 'px',
      'thumb_width' => '80',
      'thumb_border_color' => '#007bff',
    ] + parent::defaultSettings();
  }

  /**
   * Settings form.
   *
   * @param array $form
   *   Form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state array.
   *
   * @return mixed
   *   Returns mixed data.
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $image_styles = image_style_options(FALSE);

    $description_link = Link::fromTextAndUrl(
      $this->t('Configure Image Styles'),
      Url::fromRoute('entity.image_style.collection')
    );

    $image_style_large_description = $this->t('Select the image style to be applied on large screens (e.g., desktops). This may include resolution, layout, or styling presets.');

    $element['image_style_large'] = [
      '#title' => $this->t('Image Style for Large Devices'),
      '#type' => 'select',
      '#default_value' => $this->getSetting('image_style_large'),
      '#empty_option' => $this->t('None (original image)'),
      '#options' => $image_styles,
      '#description' => [
        ['#markup' => $image_style_large_description],
        $description_link->toRenderable(),
      ],
      '#access' => $this->currentUser->hasPermission('administer image styles'),
    ];

    $element['image_style_medium'] = [
      '#title' => $this->t('Image Style for Medium Devices'),
      '#type' => 'select',
      '#default_value' => $this->getSetting('image_style_medium'),
      '#empty_option' => $this->t('None (original image)'),
      '#options' => $image_styles,
      '#description' => $this->t('Choose the image style for medium-sized devices such as tablets or small laptops.'),
      '#access' => $this->currentUser->hasPermission('administer image styles'),
    ];

    $element['image_style_medium'] = [
      '#title' => $this->t('Image Style for Medium Devices'),
      '#type' => 'select',
      '#default_value' => $this->getSetting('image_style_medium'),
      '#empty_option' => $this->t('None (original image)'),
      '#options' => $image_styles,
      '#description' => $this->t('Choose the image style for medium-sized devices such as tablets or small laptops.'),
      '#access' => $this->currentUser->hasPermission('administer image styles'),
    ];

    $element['image_style_small'] = [
      '#title' => $this->t('Image Style for Small Devices'),
      '#type' => 'select',
      '#default_value' => $this->getSetting('image_style_small'),
      '#empty_option' => $this->t('None (original image)'),
      '#options' => $image_styles,
      '#description' => $this->t('Define the image style optimized for mobile or small screen devices.'),
      '#access' => $this->currentUser->hasPermission('administer image styles'),
    ];

    $element['thumbnail_image_style'] = [
      '#title' => $this->t('Thumbnail Image Style'),
      '#type' => 'select',
      '#default_value' => $this->getSetting('thumbnail_image_style'),
      '#empty_option' => $this->t('None (original image)'),
      '#options' => $image_styles,
      '#description' => $this->t('Define the visual style or preset applied to thumbnail images'),
      '#access' => $this->currentUser->hasPermission('administer image styles'),
    ];

    $element['thumbnail_border_style'] = [
      '#title' => $this->t('Thumbnail Border Style'),
      '#type' => 'select',
      '#default_value' => $this->getSetting('thumbnail_border_style'),
      '#options' => [
        'round' => $this->t('Round'),
        'square' => $this->t('Square'),
      ],
      '#description' => $this->t('Choose the border shape for thumbnails. Common options include square or round.'),
      '#access' => $this->currentUser->hasPermission('administer image styles'),
    ];

    $element['allow_thumbnail_overlap'] = [
      '#title' => $this->t('Allow Overlapping Effect for Thumbnails'),
      '#type' => 'checkbox',
      '#default_value' => $this->getSetting('allow_thumbnail_overlap') ?? TRUE,
      '#description' => $this->t('Enable a visual overlap effect between adjacent thumbnails for a stylized layout.'),
    ];

    $element['thumbnail_overlap_threshold'] = [
      '#title' => $this->t('Apply Overlap Only If Thumbnails Exceed'),
      '#type' => 'number',
      '#min' => '0',
      '#max' => '100',
      '#step' => '1',
      '#default_value' => $this->getSetting('thumbnail_overlap_threshold') ?? '5',
      '#description' => $this->t('Set the minimum number of thumbnails required before the overlapping effect is applied. Helps control layout clutter.'),
      '#states' => [
        'visible' => [
          ':input[name$="[allow_thumbnail_overlap]"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $element['enable_mouse_zoom'] = [
      '#title' => $this->t('Enable Zoom on Mouse Wheel'),
      '#type' => 'checkbox',
      '#default_value' => $this->getSetting('enable_mouse_zoom') ?? TRUE,
      '#description' => $this->t('Allow users to zoom in and out of the image using the mouse wheel for better inspection.'),
    ];

    $element['default_zoom_scale'] = [
      '#title' => $this->t('Default Zoom Scale'),
      '#type' => 'number',
      '#min' => '0',
      '#max' => '10',
      '#step' => '1',
      '#default_value' => $this->getSetting('default_zoom_scale') ?? '2',
      '#description' => $this->t('Define the initial zoom scale when the image is loaded. A value of 1 represents original size.'),
    ];
    $element['show_magnifier'] = [
      '#title' => $this->t('Show Magnifier'),
      '#type' => 'checkbox',
      '#default_value' => $this->getSetting('show_magnifier') ?? TRUE,
      '#description' => $this->t('Display a magnifier tool when hovering over the image, providing a close-up view of the hovered area.'),
    ];

    $element['image_class'] = [
      '#title' => $this->t('Image class'),
      '#type' => 'textfield',
      '#description' => $this->t('Custom class for the image'),
      '#size' => '20',
      '#default_value' => $this->getSetting('image_wrapper_class'),
    ];

    $element['image_wrapper_class'] = [
      '#title' => $this->t('Wrapper class'),
      '#type' => 'textfield',
      '#description' => $this->t('Custom class for the image wrapper'),
      '#size' => '20',
      '#default_value' => $this->getSetting('image_wrapper_class'),
    ];

    $element['container_width'] = [
      '#title' => $this->t('Container Max Width'),
      '#type' => 'number',
      '#min' => '0',
      '#step' => '1',
      '#default_value' => $this->getSetting('container_width') ?? '600',
      '#description' => $this->t('Specify the container max width. Used together with the selected unit.'),
    ];

    $element['container_width_unit'] = [
      '#title' => $this->t('Container Width Unit'),
      '#type' => 'select',
      '#options' => [
        'px' => $this->t('Pixels (px)'),
        'em' => $this->t('EM (em)'),
        'rem' => $this->t('REM (rem)'),
        '%' => $this->t('Percent (%)'),
      ],
      '#default_value' => $this->getSetting('container_width_unit') ?? 'px',
      '#description' => $this->t('Select the CSS unit for the container max width.'),
    ];

    $element['thumb_width'] = [
      '#title' => $this->t('Thumbnail Width'),
      '#type' => 'select',
      '#options' => array_combine(
      range(20, 140, 10),
      array_map(function ($value) {
        return $this->t('@value px', ['@value' => $value]);
      }, range(20, 140, 10))
      ),
      '#default_value' => $this->getSetting('thumb_width') ?? 80,
      '#description' => $this->t('Choose the width (in pixels) to apply to each thumbnail.'),
    ];

    $element['thumb_border_color'] = [
      '#title' => $this->t('Thumbnail Border Color'),
      '#type' => 'textfield',
      '#default_value' => $this->getSetting('thumb_border_color') ?? '#007bff',
      '#description' => $this->t('Select the color for the border around each thumbnail. You can use any valid CSS color format, including hexadecimal (e.g., #ff0000), RGB (e.g., rgb(255, 0, 0)), HSL (e.g., hsl(0, 100%, 50%)), or named colors (e.g., red).'),
    ];

    return $element;
  }

  /**
   * {@inheritdoc}
   */

  /**
   * Settings summary.
   *
   * @return array
   *   Summary of settings.
   */
  public function settingsSummary() {
    $summary = [];

    // Image Style for Large Devices.
    $image_styles = image_style_options(FALSE);
    unset($image_styles['']);
    $image_style_large = $this->getSetting('image_style_large');
    if (isset($image_styles[$image_style_large])) {
      $summary[] = $this->t('Image Style for Large Devices: @value', ['@value' => $image_styles[$image_style_large]]);
    }
    else {
      $summary[] = $this->t('Image Style for Large Devices: None (original image)');
    }

    // Image Style for Medium Devices.
    $image_style_medium = $this->getSetting('image_style_medium');
    if (isset($image_styles[$image_style_medium])) {
      $summary[] = $this->t('Image Style for Medium Devices: @value', ['@value' => $image_styles[$image_style_medium]]);
    }
    else {
      $summary[] = $this->t('Image Style for Medium Devices: None (original image)');
    }

    // Image Style for Small Devices.
    $image_style_small = $this->getSetting('image_style_small');
    if (isset($image_styles[$image_style_small])) {
      $summary[] = $this->t('Image Style for Small Devices: @value', ['@value' => $image_styles[$image_style_small]]);
    }
    else {
      $summary[] = $this->t('Image Style for Small Devices: None (original image)');
    }

    // Thumbnail Image Style.
    $thumbnail_image_style = $this->getSetting('thumbnail_image_style');
    if (isset($image_styles[$thumbnail_image_style])) {
      $summary[] = $this->t('Thumbnail Image Style: @value', ['@value' => $image_styles[$thumbnail_image_style]]);
    }
    else {
      $summary[] = $this->t('Thumbnail Image Style: None (original image)');
    }

    // Thumbnail Border Style.
    $thumbnail_border_style = $this->getSetting('thumbnail_border_style');
    $summary[] = $this->t('Thumbnail Border Style: @value', ['@value' => $thumbnail_border_style]);

    // Allow Thumbnail Overlap.
    $allow_thumbnail_overlap = $this->getSetting('allow_thumbnail_overlap') ? 'Yes' : 'No';
    $summary[] = $this->t('Allow Thumbnail Overlap: @value', ['@value' => $allow_thumbnail_overlap]);

    // Thumbnail Overlap Threshold.
    $thumbnail_overlap_threshold = $this->getSetting('thumbnail_overlap_threshold');
    $summary[] = $this->t('Thumbnail Overlap Threshold: @value', ['@value' => $thumbnail_overlap_threshold]);

    // Enable Mouse Zoom.
    $enable_mouse_zoom = $this->getSetting('enable_mouse_zoom') ? 'Enabled' : 'Disabled';
    $summary[] = $this->t('Mouse Zoom: @value', ['@value' => $enable_mouse_zoom]);

    // Default Zoom Scale.
    $default_zoom_scale = $this->getSetting('default_zoom_scale');
    $summary[] = $this->t('Default Zoom Scale: @value', ['@value' => $default_zoom_scale]);

    // Show Magnifier.
    $show_magnifier = $this->getSetting('show_magnifier') ? 'Yes' : 'No';
    $summary[] = $this->t('Show Magnifier: @value', ['@value' => $show_magnifier]);

    // Image Class.
    $image_class = $this->getSetting('image_class');
    $summary[] = $this->t('Image class: @value', ['@value' => $image_class]);

    // Image Wrapper Class.
    $image_wrapper_class = $this->getSetting('image_wrapper_class');
    $summary[] = $this->t('Image Wrapper Class: @value', ['@value' => $image_wrapper_class]);

    // Container Width and Unit.
    $container_width = $this->getSetting('container_width');
    $container_width_unit = $this->getSetting('container_width_unit');
    $summary[] = $this->t('Container Max Width: @width @unit',
      ['@width' => $container_width, '@unit' => $container_width_unit]);

    // Thumbnail Width.
    $thumb_width = $this->getSetting('thumb_width');
    $summary[] = $this->t('Thumbnail Width: @value px', ['@value' => $thumb_width]);

    // Thumbnail border color.
    $thumb_border_color = $this->getSetting('thumb_border_color');
    $summary[] = $this->t('Thumbnail border color code: @value', ['@value' => $thumb_border_color]);

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $field_name = $items->getFieldDefinition()->getName();
    $elements = [];
    $files = $this->getEntitiesToView($items, $langcode);

    if (empty($files)) {
      return $elements;
    }

    $images = [];
    $original_urls = [];
    $cache_tags = [];

    if ($files) {
      foreach ($files as $delta => $file) {
        $image_uri = $file->getFileUri();
        $image_target_id = $file->id();

        $image_uri_absolute = $this->fileUrlGenerator->generateAbsoluteString($image_uri);
        $image_uri_parsed = parse_url($image_uri_absolute);
        $image_uri_path = $image_uri_parsed['path'] ?? '';

        $original_urls[$image_target_id] = $image_uri_path;

        $image_styles = [
          'large' => $this->getSetting('image_style_large'),
          'medium' => $this->getSetting('image_style_medium'),
          'small' => $this->getSetting('image_style_small'),
          'thumbnail' => $this->getSetting('thumbnail_image_style'),
        ];

        $styled_urls = [];
        foreach ($image_styles as $key => $style_name) {
          if (!empty($style_name)) {
            $style = $this->imageStyleStorage->load($style_name);
            if ($style) {
              $styled_urls[$key] = $style->buildUrl($image_uri);
              $cache_tags = Cache::mergeTags($cache_tags, $style->getCacheTags());
            }
            else {
              // Fallback to original URL if style cannot be loaded.
              $styled_urls[$key] = $image_uri_path;
            }
          }
          else {
            // No style selected: use original image URI.
            $styled_urls[$key] = $image_uri_path;
          }
        }

        $cache_tags = Cache::mergeTags($cache_tags, $file->getCacheTags());

        $item_attributes = [];
        $item_attributes['class'][] = 'xzoom';
        $item_attributes['xoriginal'] = $image_uri_path;
        $item_attributes['fid'] = $image_target_id;

        $images[$delta] = [
          'original_url' => $image_uri_path,
          'styled_urls' => $styled_urls,
          'attributes' => $item_attributes,
        ];
      }

      $data_attributes = [
        'thumbnail_border_style' => $this->getSetting('thumbnail_border_style'),
        'allow_thumbnail_overlap' => $this->getSetting('allow_thumbnail_overlap'),
        'thumbnail_overlap_threshold' => $this->getSetting('thumbnail_overlap_threshold'),
        'enable_mouse_zoom' => $this->getSetting('enable_mouse_zoom'),
        'default_zoom_scale' => $this->getSetting('default_zoom_scale'),
        'show_magnifier' => $this->getSetting('show_magnifier'),
        'image_class' => $this->getSetting('image_class'),
        'image_wrapper_class' => $this->getSetting('image_wrapper_class'),
        'container_width' => $this->getSetting('container_width'),
        'container_width_unit' => $this->getSetting('container_width_unit'),
        'thumb_width' => $this->getSetting('thumb_width'),
        'thumb_border_color' => $this->getSetting('thumb_border_color'),
      ];

      $elements[] = [
        '#theme' => 'product_gallery',
        '#images' => $images,
        '#data_attributes' => $data_attributes,
        '#cache' => [
          'tags' => $cache_tags,
        ],
        'field_name' => [$field_name],
        '#attached' => [
          'drupalSettings' => [
            'imageFieldZoom' => [
              'enable_mouse_zoom' => $this->getSetting('enable_mouse_zoom'),
              'default_zoom_scale' => $this->getSetting('default_zoom_scale'),
              'show_magnifier' => $this->getSetting('show_magnifier'),
              'allow_thumbnail_overlap' => $this->getSetting('allow_thumbnail_overlap'),
              'thumbnail_overlap_threshold' => $this->getSetting('thumbnail_overlap_threshold'),
              'image_urls' => $original_urls,
            ],
          ],
        ],
      ];
    }

    return $elements;
  }

}
